\\ Copyright 2021, 2022 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\-----------------------------------------------------------------------------
\\ Kevin Ryde, November 2021
\\
\\ Usage: gp <a230979.gp
\\
\\ Find the linear recurrence order for A230979, being the decimal digits of
\\ the rational
\\
\\     149597870700 / 299792458 = 499.00478383...
\\
\\ The answer is the printed "linrec_order = 440008".
\\
\\ The generating function and minimum recurrence for the digits of a
\\ fraction are normally a simple calculation from the periodic part.
\\ But here period length 880014 is a little too big for PARI's generic
\\ polynomial gcd().  So some ad-hoc shortcuts to reach the answer then
\\ verify the answer with a full check.


\\ try to quiet some increase size messages by pre-allocating
default(parisizemax, 500*10^6);  \\ 500 megabytes
default(parisize,    500*10^6);
default(recover,0);
default(strictargs,1);

A230979_data = [4,9,9,0,0,4,7,8,3,8,3,6,1,5,6,4,1,1,9,1,3,4,7,1,1,5,2,0,9,9,4,9,6,7,7,9,2,0,1,8,3,0,3,5,4,2,5,1,2,7,3,3,9,2,6,0,8,1,6,2,2,7,7,3,8,4,5,7,6,4,9,9,2,5,2,6,9,3,0,0,8,0,4,0,9,1,6,0,9,2,6,9,2,3,6,5,1,9,6,1,9,1,8,2,6,8];

\\ rat = A230979 rational value
rat = 149597870700 / 299792458;

\\ integer part
intpart = floor(rat);
intpart == 499 || error;

\\ fraction part
fra = frac(rat);
print("fra = ",fra);

num = numerator(fra);
den = denominator(fra);
rat == intpart + num/den  || error();
num == 9823               || error();
den == 2053373            || error();

\\ "maxlen" maximum possible period length is the multiplicative order
\\ of 10 mod den.  Division 10^maxlen * num / den is some quotient digits
\\ and remainder 10^maxlen*num mod den.  But 10^maxlen == 1 mod den means
\\ this remainder is back at num again and so repeat.
\\
maxlen = znorder(Mod(10,den));
Mod(10,den)^maxlen == 1  || error();

\\ digs[] is the periodic digits of fra.  They're found by a generic (but
\\ not particularly efficient) fraction to digits by successive 10*n for
\\ each digit and watch for a repeat of n meaning the pattern repeats.
\\ Here fra is purely periodic and turns out to be the whole maxlen so in
\\ fact seen[] is not needed.
\\
print("digs[] digits ...");
digs = vector(maxlen);
{
  my(n=num,
     seen=vectorsmall(den));  \\ various remainders mod den
  for(i=1,#digs,
     seen[n] = i;  \\ where n previously seen
     my(t); [t,n] = divrem(10*n,den);
     digs[i] = t;
     if(seen[n],
        seen[n] == 1  || error();  \\ seen at i=1 means purely periodic
        n == num      || error();  \\ back at original num
        digs=Vec(digs,i);  \\ length i
        break));
}
{
  print1("  = ");
  for(i=1,20, print1(digs[i],","));
  print();
}

\\ period length of the digits of fra
len = #digs;
print("period len = ",len);
len == 880014   || error();

\\ actual len is the maxlen in fact
len == maxlen   || error();

\\ intpart=499 then digs[] gives A230979_data[]
{ my(got=concat(digits(intpart),digs));
  A230979_data == got[1..#A230979_data] || error();
}

\\ digs / 99..99 is fra
fromdigits(digs) / (10^len - 1) == fra  || error();

\\ generating function for fra digits will be gnum / gden
gnum = Polrev(digs);
gden = 1 - x^len;
poldegree(gnum) == len - 1  || error();
poldegree(gden) == len      || error();
digs[#digs] != 0  || error();  \\ so high term x^(len-1)

\\ Normally a direct division g = gnum/gden is enough to get g in least
\\ terms and hence minimum linear recurrence order.  But here PARI's generic
\\ integer polynomial gcd() runs rather slowly.  Presumably this is
\\ expression swell in the intermediate parts of the calculation (until the
\\ final coeffs are 1s actually).
\\ 
\\ A shortcut is to get a prospective gcd over coefficients mod 2, which
\\ runs much faster.  It turns out to be the full answer.  As a strategy
\\ this is likely to be less general than it sounds, but works here.

digs[#digs] == 9  || error();   \\ is odd

print("gcd gnum,gden over GF(2)[x] ...");
gf2_gcd = lift(gcd(Mod(gnum,2), Mod(gden,2)));

\\ gf2_gcd is the following half length factor
print("compare 1 + x + ... + x^(halflen-1) ...");
halflen = len / 2;
poldegree(gf2_gcd) == halflen-1        || error();
gf2_gcd == (1-x^halflen)/(1-x)         || error();  \\ expression
gf2_gcd == Pol(vector(halflen,i, 1))   || error();  \\ all 1s

\\ gf2_gcd has to be thought of as only potentially a divisor of gnum and
\\ gden.  For gden, it's easily seen to be a divisor algebraically since len
\\ is even and so gden is the difference of two perfect squares.  For gnum,
\\ it's not so clear, but doing the division shows it's right.
\\
gden_q = (1+x^halflen) * (1-x);
print("gden_q = ",gden_q);
gden_q == 1 - x + x^440007 - x^440008  || error();

\\ gden_q is verified by multiplying back
print("gden = gden_q * gf2_gcd multiply check ...");
gden == gden_q * gf2_gcd  || error();

\\ General purpose divrem(gnum,gf2_gcd) is a little on the slow side, so
\\ take a shortcut by
\\
\\     gnum / gf2_gcd  =  gnum*(1-x) / (1-x^halflen)
\\
\\ The following is an unsophisticated sparse divide by 1-x^halflen.
\\ It's nearly as easy as quotient = negative of the high half coeffs, but
\\ gnum*(1-x) is degree len so there's one coefficient worth of overlap
\\ where x^len reduces to x^halflen and in turn reduces to the constant term
\\ position.
{
  my(v=Vecrev(gnum*(1-x)));   \\ coeffs from constant upwards
  my(qv=vector(#v-halflen),   \\ quotient coeffs
     upto=#qv+1);
  forstep(hi=#v,1,-1,
     my(lo = hi-halflen);
     if(lo<1,break);
     f = -v[hi];        \\ 1-x^halflen with x^halflen at v[hi]
     v[lo] -= f*(1);    \\ multiple of that to clear v[hi]
     v[hi] -= f*(-1);
     qv[upto--] = f);

  upto == 1 || error();
  gnum_q = Polrev(qv);  \\ quotient
  gnum_r = Polrev(v);   \\ remainder

  \\ remainder 0, in all coeffs, means the division was exact
  gnum_r == 0  || error();
}
print("gnum_q degree ",poldegree(gnum_q));

\\ gnum_q was set by a shortcut and is verified by multiplying back.
print("gnum = gnum_q * gf2_gcd multiply check ...");
gnum == gnum_q * gf2_gcd  || error();   \\ gf2_gcd is a divisor of gnum

\\ g = gnum_q/gden_q division is much faster than full gnum/gden.  It turns
\\ out gf2_gcd is not just a divisor of gnum and gden, but is their gcd so
\\ 
\\     gcd(gnum_q,gden_q) == 1
\\     
\\ This is detected reasonably quickly in the ordinary gnum_q/gden_q (but
\\ not instantly), so that the resulting g is as easy as gnum_q and gden_q
\\ as numerator and denominator.
\\
print("g = gnum_q/gden_q ...");
g = gnum_q / gden_q;     \\ t_RFRAC of the fraction part gf

numerator(g)   == gnum_q       || error();
denominator(g) == gden_q       || error();
denominator(g) == 1 - x + x^440007 - x^440008   || error();

\\ Linear recurrence order is the degree of the denominator of g, in the
\\ usual way.
\\
linrec_order = poldegree(denominator(g));
print("linrec_order = ",linrec_order);

linrec_order == 440008       || error();
linrec_order == halflen + 1  || error();


\\ check that g as a generating function makes some repetitions of digs[]
print("g series check ...");
{ my(three_digs=concat([digs,digs,digs]));
  Ser(g,,#three_digs) == Ser(three_digs)  || error();
}

\\ Substituting 1/10 into g is exactly fra again, and which is the whole
\\ purpose of making such g, and is a final check.  Note that g has the
\\ first digit of fra as the x^0 constant term and this is to be (1/10)^1
\\ of the fraction, and so on further digits, and so substitute into g*x.
\\
\\ This subst may take a little time to run.  Suspect this is large-ish
\\ intermediate rationals until the modest size final "fra".
\\
print("subst g*x is fra again (constant term gets 1/10) ...");
subst(g*x, x, 1/10) == fra  || error();

print("end");
