\\ Copyright 2019, 2020, 2023 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


default(strictargs,1);
default(recover,0);
read("../vecper.gp");
read("../test-helpers.gp");
read("OEIS-data.gp");
read("OEIS-data-wip.gp");

option_verbose = 0;

\\-----------------------------------------------------------------------------
to_binary(n)=fromdigits(binary(n))*sign(n);
from_binary(n)=fromdigits(digits(n),2);
is_integer(x) = (x==floor(x)) && variable(x)==0;
is_power_of_2(x) = is_integer(x) && x>=1 && valuation(x,2)==logint(x,2);

bitlength(n) = if(n, logint(n,2)+1, 0);
\\ but A070939 takes one leading 0 bit on 0, unlike all other numbers
OEIS_check_func("A070939",(n)->if(n==0,1,bitlength(n)));

high_bit_k(n) = logint(n,2);
{
  OEIS_check_func("A000523","high_bit_k");
  check_equal(high_bit_k(7), 2);
  check_equal(high_bit_k(8), 3);
  check_equal(high_bit_k(9), 3);
  check_equal(vector(1000,n, my(k=high_bit_k(n)); 2^k <= n && n < 2^(k+1)),
              vector(1000,n, 1));
  \\ A000523 floor(log2(n))
}

high_bit(n) = if(n==0,0, 1<<high_bit_k(n));
OEIS_check_func("A053644","high_bit");


\\-----------------------------------------------------------------------------

\\ period length of the repeating part of fraction f in binary
frac_periodic_length(f) = \
  vecper_periodic_length(vecper_frac_digits(f,2));
{
  \\ A007733 period of 1/n in binary, multiplicative order of 2 mod n
  OEIS_check_func("A007733",(n)->frac_periodic_length(1/n));

  check_equal(vector(1000,n, frac_periodic_length(1/n)),
              vector(1000,n, znorder(Mod(2,n>>valuation(n,2)))));
}
{
  \\ A002326 period of 1/(2n+1) in binary
  OEIS_check_func("A002326",(n)->frac_periodic_length(1/(2*n+1)));

  check_equal(vector(100,n,n--; frac_periodic_length(1/(2*n+1))),
              vector(100,n,n--; znorder(Mod(2,2*n+1))));
}
{
  \\ 1/odd from the first 1-bit is purely periodic
  check_equal(vector(500,n, n=2*n-1;
                     my(v=vecper_frac_digits(1/n,2));
                     v=vecper_normalize(v);
                     vecper_fixed_length(v)==0),
              vector(500,n, 1),
              "1/odd purely periodic");
  \\ forstep(n=1,20,2, \
  \\   my(n=2*n-1); \
  \\   print(vecper_periodic(vecper_frac_digits(1/n,2))));

  \\ Repeated *2 and mod n (odd)
  \\ 2^k = 1 mod n eventually
  \\ because finite number of residues 2^k mod n
  \\ so eventually cyclic so 2^k mod n = 2^j mod n is 2^(k-j) = 1 mod n
}

\\ in binary
frac_periodic_part(f) =
{
  (0 <= f && f <= 1) || error();
  if(f==1,return(0));
  \\ while(2*numerator(f) < denominator(f), f*=2);
  fromdigits(vecper_periodic(vecper_frac_digits(f,2)),
             2);
}
{
  check_equal(vecper_frac_digits(1,2), [[], [1]]);
  check_equal(frac_periodic_part(1), 0);
  check_equal(frac_periodic_part(1/2), 0);
  check_equal(frac_periodic_part(1/3), 1);  /* 01 */
  check_equal(frac_periodic_part(2/3), 2);  /* 10 */
}
\\ for(n=1,20, \
\\   my(n=2*n-1); \
\\   print1(frac_periodic_part(1/n)","));
\\
\\   print(n" "vecper_frac_digits(1/n,2)));
\\   print1(to_binary(per),","));
{
  \\ periodic part of 1/odd (purely periodic with log2 leading 0s)
  \\ 1,3,1,7,93,315,1,15,13797,3,89,41943,9709,9256395,1,31,117,1857283155,105
  \\ A165781 = (2^A002326(n)-1)/(2*n+1).

  OEIS_check_func("A165781",(n)->if(n==0,1, frac_periodic_part(1/(2*n+1))));
}

\\ periodic part of 1/odd in binary
\\ not in OEIS:    1, 11, 1, 111, 1011101, 100111011, 1,  1111, 11010111100101, 11, 1011001,1010001111010111,10010111101101
\\ cf len       1, 2,  4, 3, 6,   10,      12,        4,  8,    18
\\                 3   5  7  9    11       13         15  17    19
\\ vector(12,n,znorder(Mod(2,2*n-1))) =
\\
\\ periodic part of 1/any bits
\\ not in OEIS: 1,1,2,1,12,2,4,1,56,12,744,2,2520,4,8,1,240,56,220752,12
\\ periodic part of 1/any coded into binary
\\ not in OEIS: 1,1,10,1,1100,10,100,1,111000,1100,1011101000,10,100111011000,100,1000

\\ znorder requires base,modulus coprime,
\\ so must divide out n factors of 2
periodic_part_compact(n) = {     \\ any n
  n>>=valuation(n,2);
  if(n==1,0, (2^znorder(Mod(2,n)) - 1)/n);
  \\ logint(n,2));
}
{
  check_equal(vector(500,n, periodic_part_compact(n)),
              vector(500,n, frac_periodic_part(1/n)),
              "periodic_part_compact()")
}

  \\ n=2*n+1; my(s=1<<logint(n,2), p=s, l=List([]));
  \\ until(p==s, p*=2; listput(l,p>=n); if(p>=n,p-=n));
  \\ fromdigits(Vec(l),2);

odd_periodic_part_compact(n) = {
  n=2*n+1; bitneg(0,znorder(Mod(2,n))) / n;
}
{
  check_equal(vector(50,n, frac_periodic_part(1/(2*n+1))),
              vector(50,n, odd_periodic_part_compact(n)),
              "odd_periodic_part_compact()");
  OEIS_check_func("A165781","odd_periodic_part_compact");
}

\\ frac_periodic_part(1/19)
\\ odd_periodic_part_compact(19)
\\ to_binary(odd_periodic_part_compact(9))
\\ to_binary(frac_periodic_part(1/(2*9+1)))
\\ 2*12 - 19 == 5

\\ vector(10,n, my(n=2*n+1); (2^n - 1) / n
{
  \\ recovering the fraction

  check_equal(vector(1000,n, my(n2 = 2*n+1);
                     my(b=odd_periodic_part_compact(n),
                        L=logint(n2,2),         \\ initial 0s
                        w=bitlength(b) + L);    \\ total width
                     w == frac_periodic_length(1/n2) || error();
                     b / (2^w - 1)),
              vector(1000,n, my(n2=2*n+1);
                     1/n2),
              "odd_periodic_part_compact() recover 1/n");

  \\ Wrongish:
  \\ check_equal(vector(10,n, my(b=odd_periodic_part_compact(n));
  \\                    b / (2^bitlength(b)-1)),
  \\             vector(10,n, my(n=2*n+1); n),
  \\             "odd_periodic_part_compact() recover n");
  \\ check_equal(vector(10,n, my(b=odd_periodic_part_compact(n));
  \\                    b / (2^bitlength(b)-1)),
  \\             vector(10,n, 1/(2*n+1) * high_bit(2*n+1)));
  \\ check_equal(vector(10,n, my(b=odd_periodic_part_compact(n));
  \\                    b / (2^bitlength(b)-1)),
  \\             vector(10,n, 1/(2*n+1) * high_bit(2*n+1)));
}

{
  \\ A160144
  OEIS_check_func("A160144",
                  (n)->my(n=2*n+1); numerator(n / (2^n - 1)));
  \\ apply(to_binary,want)
  \\ not in OEIS: 1,11,101,111,1001,1011,1101,1111,10001,10011,11,10111,11001
}
{
  \\ A303449
  OEIS_check_func("A303449",
                  (n)->my(n=2*n+1); denominator(n / (2^n - 1)));
  \\ apply(to_binary,OEIS_data("A303449"))
  \\ not in OEIS: 1, 111, 11111, 1111111, 111111111, 11111111111, 1111111111111, 111111111111111, 11111111111111111, 1111111111111111111, 1001001001001001001, 11111111111111111111111, 1111111111111111111111111
}

\\-----------------------------------------------------------------------------
\\ binary fraction, periodic part starting with the first 1-bit
periodic_part_1n_start1(n) = frac_periodic_part(high_bit(n)/n);

{
  \\ period length unchanged by mult up 2^k * 1/n
  check_equal(vector(500,n, vecper_periodic_length(vecper_frac_digits(high_bit(n)/n,2))),
              vector(500,n, vecper_periodic_length(vecper_frac_digits(1/n,2))));

  \\ starting 1
  check_equal(vector(500,n, vecper_periodic(vecper_frac_digits(high_bit(n)/n,2))[1]),
              vector(500,n, 1));

  check_equal(vector(1000,n, bitlength(periodic_part_1n_start1(n))),
              vector(1000,n, if(is_power_of_2(n),0,znorder(Mod(2,n>>valuation(n,2))))));
}

quit

myord(b,m) = m /= b^valuation(m,b); znorder(Mod(b,m));
{
  check_equal(vector(1000,n, my(p=periodic_part_1n_start1(n)); 
                     p/(2^myord(2,n)-1) / high_bit(n)),
              vector(1000,n, if(is_power_of_2(n),0, 1/n)));
}
quit

% b=frac_periodic_part(1/(2*9+1))
% b=odd_periodic_part_compact(9)
% to_binary(b)
% l=bitlength(b)
% b/(2^l-1)
% b/(2^l-1) = 2^k/odd
% b = 2^k * (2^l-1) / odd
% vector(10,n, odd_periodic_part_compact(n))
% vector(10,n, my(n=2*n+1); numerator((2^n - 1) / n))
% vector(10,n, my(n=2*n+1); numerator((2^n - 1) / n) * high_bit(n))

% forstep(n=1,200,2, \
%   my(b = numerator((2^n - 1) / n)); \
%   if(is_power_of_2(b+1),next); \
%   print(to_binary(b)));

% forstep(n=1,20,2, \
%   print1(periodic_part_1n_start1(n),","));
%   print1(to_binary(periodic_part_1n_start1(n)),","));
%   my(n=2*n-1); \
%
% odds (nothing from superseeker)
% not in OEIS: 12,56,2520,240,48,671088,148102320,992,59433060960,
% not in OEIS: 1100,111000,100111011000,11110000,110000,10100011110101110000,1000110100111101110010110000
%
% all
% not in OEIS: 2,12,4,56,744,2520,8,240,220752
% not in OEIS: 10,1100,100,111000,1011101000,100111011000,1000,11110000,110101111001010000

% #vecper_frac_digits(1/2^10,3)[2]
% znorder(Mod(2,3^10))
% znorder(Mod(3,2^50))
