\\ Copyright 2019, 2020, 2022, 2023 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Usage: read("vecper.gp")
\\
\\ This is some code representing and acting on periodic or eventually
\\ periodic infinite vectors, comprising an initial fixed part and a
\\ periodic part.  addhelp() strings have docs of each function and see
\\ notes in the comments below.
\\
\\     vec = vecper_fixed(v)                    fixed part
\\     len = vecper_fixed_length(v)
\\     vec = vecper_periodic(v)                 periodic part
\\     len = vecper_periodic_length(v)
\\
\\     elem = vecper_get(v,i)                   element i=1,2,3,...
\\     vec = vecper_get_range(v,lo,hi)
\\     poly = vecper_polrev(v,var='x)
\\
\\     order = vecper_lex(u,v)
\\     bool = vecper_is_equal(u,v)
\\
\\     v = vecper_make_state(func,state=0)
\\     v = vecper_apply(func,v)
\\     v = vecper_apply_state(func,v,state=0)
\\     v = vecper_select(func,v)
\\     v = vecper_normalize(v)
\\
\\     v = vecper_frac_digits(f,base=10,flag=0)
\\     f = vecper_frac_fromdigits(v,base=10)
\\
\\
\\ Periodic Vector
\\ --------------
\\
\\ A vecper is a vector of 2 vectors: a fixed part vector and a periodic
\\ part vector.  The accessors vecper() and vecper_periodic() return each.
\\
\\     v = [ [1,2], [3,4,5] ]
\\     represents 1,2, 3,4,5, 3,4,5, 3,4,5, ...
\\
\\ vecper represents an "infinite" vector, but one which is eventually
\\ periodic so it is a pair of finite vectors.  If the periodic part is
\\ empty then the whole vecper is finite.  If both fixed and periodic are
\\ empty then the vecper is empty.
\\
\\     v = [ [1,2,3], [] ]          empty periodic part
\\     v = [ [], [] ]               both parts empty
\\
\\ Vector elements can be any types and any mixture of types, the same as
\\ ordinary vectors.
\\
\\ The fixed and periodic parts don't have to be in a "reduced" or minimum
\\ form.  The periodic part might be a few copies of some smaller period, or
\\ the fixed part might have a few elements which are the same as what the
\\ periodic part would give.  vecper_normalize() reduces to the minimum if
\\ desired.
\\
\\     vecper_normalize([[1,2],[3,4,3,4]]) == [[1,2],[3,4]]
\\     vecper_normalize([[1,2,1],[2,1,2,1]]) == [[],[1,2]]
\\
\\
\\ Transforming
\\ ------------
\\
\\ Some GP operators and functions act on the elements of nested vectors
\\ like a vecper.  For example,
\\
\\     [ [0,1], [1,1,0] ] * 2 == [ [0,2], [2,2,0] ]
\\     real([ [3+I], [2+2*I,5+4*I] ]) == [[3], [2,5]]
\\
\\ For other transformations, vecper_apply() is similar to plain vector
\\ apply()
\\
\\     vecper_apply(elem->elem+2, [[1],[3,2]]) == [[3],[5,4]]
\\
\\ vecper_apply_state() brings a "state" along with the transformation.
\\ The state can remember things like some preceding elements, or odd/even
\\ position in the vector, etc.  See examples/dragon-curve.gp for such a
\\ transformation.
\\
\\ The transformation functions don't try to notice possible reductions.
\\ That's left to later vecper_normalize() if you need it.  Unreduced can be
\\ good to keep a few vecpers looking similar, and usually at worst results
\\ in an lcm() of period lengths.
\\
\\
\\ Fraction Digits
\\ ---------------
\\
\\ Rationals 0 <= f <= 1 can be represented as digits with a fixed initial
\\ part and then periodic part.  vecper_frac_digits() gives the digits of f
\\ as a vecper.  vecper_frac_fromdigits() converts back to a rational.
\\
\\     1/14 = . 0 714285 714285 ...
\\     vecper_frac_digits(1/14) == [[0], [7,1,4,2,8,5]]
\\     vecper_frac_fromdigits([[0],[7,1,4,2,8,5]]) == 1/14
\\
\\ vecper_frac_digits() takes any integer base >= 2.  The converse
\\ vecper_frac_fromdigits() takes any base and any digit values (not just
\\ integers).  See examples/frac-digits.gp for some samples.
\\
\\ When going to digits and then back fromdigits, you can apply some
\\ conversions, mutations, extractions, etc, in between.  See
\\ examples/Z-order-curve.gp for a simple bit extraction.
\\
\\ If f is an exact rational n/10^k for n in the range 0 < n < 10^k then it
\\ has two representations,
\\
\\     7/10 = 0.700000...
\\          = 0.699999...         parameter flag=1
\\
\\ The flag parameter to vecper_frac_digits() can be 1 for the 9s form, and
\\ in general "base - 1" for other bases.  The two forms convert back the
\\ same in vecper_frac_fromdigits().
\\
\\     vecper_frac_digits(7/10)    == [[7], [0]]
\\     vecper_frac_digits(7/10,,1) == [[6], [9]]      \\ with "nines"
\\     vecper_frac_fromdigits([[7],[0]]) == 7/10
\\     vecper_frac_fromdigits([[6],[9]]) == 7/10
\\
\\
\\ Other Ways To Do It
\\ -------------------
\\
\\ GP type t_RFRAC effectively represents an infinite vector of polynomial
\\ coefficients which follow a linear recurrence.  Such recurrences includes
\\ periodic or eventually periodic terms.
\\
\\ RFRACs can be manipulated in various usual generating-function ways for
\\ shifts, adding, or various transformations.  vecper_polrev() can give an
\\ RFRAC if desired.  The disadvantages of RFRAC are that coefficients must
\\ be a numeric type (but including polynomials in other variables of lower
\\ "priority"), arbitrary transformations of terms is not particularly
\\ convenient, getting a particular term is some polmod powering or a long
\\ enough Ser(), and the period length may be obscured by polynomial factors
\\ cancelling between numerator and denominator.
\\
\\
\\ Changes
\\ -------
\\
\\ Version 1 - the first version.
\\ Version 2 - fix for vecper_normalize().


\\-----------------------------------------------------------------------------
\\ Internal Helpers

\\ v is a vector.
\\ Return it rotated so that the start is after pos.
\\ Ie. start at v[pos+1} so v[pos+1..#v, 1..pos].
\\ pos is interpreted mod mod #v if not already ranging 1..#v.
vecper_INTERNAL_vector_rotate_after(v,pos) = \
  if(#v<2 || !(pos%=#v), v, concat([v[pos+1..#v], v[1..pos]]));

\\ v is a vector.
\\ Return a "rotated down" [ v[2], ..., v[#v], v[1] ]
\\ where v[1] is sent down to be the last element.
\\ n is how many repeats of this, and can be negative too.
\\ n is reduced mod #v, since multiples of #v are no change.
\\ If v is smaller than 2 elements then return unchanged.
\\ vecper_INTERNAL_vector_rotate_down(v,n=1) = \
\\   if(#v<2 || !(n%=#v), v, n=#v-n; concat([v[n+1..#v], v[1..n]]));


\\-----------------------------------------------------------------------------
\\ Accessors

vecper_fixed(v) = v[1];
{
addhelp(vecper_fixed,
"vec = vecper_fixed(v)
Return a vector of the fixed part of vecper v.");
}

vecper_fixed_length(v) = #v[1];
{
addhelp(vecper_fixed_length,
"len = vecper_fixed_length(v)
Return the length of the fixed part of vecper v.");
}

vecper_periodic(v) = v[2];
{
addhelp(vecper_periodic,
"vec = vecper_periodic(v)
Return a vector of the periodic part of vecper v.");
}

vecper_periodic_length(v) = #v[2];
{
addhelp(vecper_periodic_length,
"len = vecper_periodic_length(v)
Return the length of the periodic part of vecper.");
}

vecper_get(v,i) = if(i<=#v[1], v[1][i], v[2][(i-#v[1]-1) % #v[2] + 1]);
{
addhelp(vecper_get,
"elem = vecper_get(v,i)
v is a vecper.  Return its element at index i.  The first element is i=1.
If the periodic part of v is empty then i must be within just the fixed
part.");
}

\\ Have some (much) hairier block-wise code in test.gp if wanted more speed.
\\
vecper_get_range(v,lo,hi) = vector(hi-lo+1,i, vecper_get(v,i+lo-1));
{
addhelp(vecper_get_range,
"vec = vecper_get_range(v,lo,hi)
Return a sub-vector vecper v[lo..hi] inclusive.
The return vec has length hi-lo+1.  Can have hi=lo-1 for an empty range.");
}

vecper_polrev(v,var='x) = \
  Polrev(v[1],var) + if(#v[2], Polrev(v[2],var) / (1 - var^#v[2]) * var^#v[1]);
{
addhelp(vecper_polrev,
"poly = vecper_polrev(v,var='x)
v is a vecper.  Return its elements as coefficients of a polynomial, similar
to Polrev().  v[1] is the constant term

    poly = v[1]*x^0 + v[2]*x^1 + v[3]*x^2 + ...

If the periodic part of v is empty or all 0s then this is the same as
Polrev(vecper_fixed(v))).

If the periodic part is not empty then the return is a t_RFRAC.  For example
[[4],[5,6]] is
    4 + x*(5 + 6*x)/(1-x^2)  = 4 + 5*x + 6*x^2 + 5*x^3 + 6*x^4 + ...");
}


\\-----------------------------------------------------------------------------
\\ Comparisons

vecper_INTERNAL_compare(u,v,func) =
{
  \\ print("vecper_INTERNAL_compare() "func);
  my(len,
     ret=0);  \\ return value if equal up to the shorter of the two
  if(#u[2],
     if(#v[2],
        len=max(#u[1],#v[1]) + lcm(#u[2],#v[2]),  \\ both periodic
        len=#v[1]; ret=1);                        \\ v finite
     ,
     if(#v[2],
        len=#u[1]; ret=-1,                        \\ u finite
        return(func(u[1],v[1]))));                \\ both finite
  \\ print(u" "v" "len);
  for(i=1,len,
     my(c=func(vecper_get(u,i), vecper_get(v,i)));
     if(c,return(c)));
  ret;
}

vecper_lex(u,v) = vecper_INTERNAL_compare(u,v,lex);
{
addhelp(vecper_lex,
"order = vecper_lex(u,v)
Return -1, 0 or 1 for a lex() comparison of the elements of vecpers u and v.
The return is lex(u_elem,v_elem) at the first place u,v differ, or if the
same everywhere then return 0.

If one or both u,v have empty periodic part, so a finite vector, then
elements are compared through to the end of the shortest.  If those elements
are equal then the longer vector is bigger (the same as lex() on plain
vectors).

lex() accepts various reals, strings, and further vectors, but not two
different types, so u,v must have pairwise compatible elements.");
}

vecper_is_equal(u,v) =
{
  my(len);
  if(#u[2],
     #v[2] || return(0);  \\ u infinite, v finite, is different
     len = max(#u[1],#v[1]) + lcm(#u[2],#v[2]);   \\ both infinite
     ,
     if(#v[2]                   \\ u finite, v infinite,
        || (len=#u[1])!=#v[1],  \\ or both finite but different lengths,
        return(0)));            \\ is different
  for(i=1,len,
     vecper_get(u,i)==vecper_get(v,i) || return(0));
  1;
}
{
addhelp(vecper_is_equal,
"bool = vecper_is_equal(u,v)
Return 1 if vecpers u==v, meaning each element equal.
Elements are compared per the \"==\" operator so can be any types.

If u and v are different lengths then they are not equal.
This is when either both finite and different lengths, or one finite and the
other infinite (periodic).");
}


\\-----------------------------------------------------------------------------
\\ Making

vecper_make_state(func,state=0) =
{
  \\ print("vecper_make_state()");
  my(seen=List([state]), seen_at=List([0]), l=List([]),
     ret);
  while(
        \\ print("  at state="state);
        (ret=func(state)) != 'end
        ,
        listput(l,ret[1]);
        state=ret[2];
        my(pos=setsearch(seen,state));
        if(pos,
           \\ print("  repeat state "state" pos="pos" at "seen_at[pos]);
           \\ print("  l="Vec(l));
           return(if(seen_at[pos],
                     [ Vec(l,seen_at[pos]),
                       Vec(l)[seen_at[pos]+1..#l] ];
                     ,
                     [ [], Vec(l) ])));
        pos=setsearch(seen,state,1);
        listinsert(seen,state,pos);
        listinsert(seen_at,#l,pos));

  \\ end, finite
  [Vec(l),[]];
}
{
addhelp(vecper_make_state,
"v = vecper_make_state(func,state=0)
Make a vecper by calling func and watching for repeat state.
func is called on a state and should return a 2-element vector

  [elem,new_state] = func(state)

Each elem goes into the vecper.  state is initially the given state
parameter, then each successive new_state.  state must eventually repeat.
For example a modulo counter,

  vecper_make_state((state)-> [10*state, (state+1)%3],
                    -1)
     == [[-10], [0,10,20]]

When \"state\" repeats, the elements from its previous position up to the
present position are the periodic part.  \"state\" values can be anything
and are compared by the universal cmp() (which is the strict \"===\") for
identifying a repeat.

func() can also return symbol 'end for the end of a finite vector.  The
elements so far (if any) become the finite part, and empty periodic part.");
}


\\-----------------------------------------------------------------------------
\\ Modifying

vecper_apply(func,v) =
{
  v[1]=apply(func,v[1]);
  v[2]=apply(func,v[2]);
  v;
}
{
addhelp(vecper_apply,
"v = vecper_apply(func,v)
Apply func to each element of v and return the result as a new vecper.
func is a function called as new_elem = func(elem).");
}

vecper_apply_state(func,v,state=0) =
{
  \\ print("vecper_apply_state() "v);
  for(i=1,#v[1],
     my(ret=func(v[1][i],state));
     v[1][i]=ret[1];
     state  =ret[2]);
  \\ print("  transformed v[1] = "v[1]"  state="state);
  my(vecs=vecper_make_state((state)->
                            my(vec=vector(#v[2]));
                            for(i=1,#vec,
                               my(ret=func(v[2][i],state));
                               vec[i]=ret[1];
                               state =ret[2]);
                            [vec,state];
                            ,
                            state));
  if(#vecs[1],
     v[1]=concat(v[1],concat(vecs[1][1])));
  v[2]=concat(vecs[2]);
  v;
}
{
addhelp(vecper_apply_state,
"v = vecper_apply_state(func,v,state=0)
Apply func to each element of v with a \"state\" maintained, and return the
result as a new vecper.

func is called on successive elements of v from i=1 upwards.  It should
return a 2-element vector

  [new_elem,new_state] = func(elem,state)

state is initially the given state parameter, then each successive
new_state.  For example, to negate every second element of v,

  vecper_apply_state((elem,state)-> elem*=state; state*=-1; [elem,state],
                     [[4,5],[6,7,8]],
                     1)
    == [[4,-5], [6,-7,8,-6,7,-8]]

When the \"state\" repeats after the periodic part, the new elements from
its previous occurrence up to here are the new periodic part.  \"state\"
values can be anything, but must eventually repeat over the periodic part.
States are compared by the universal cmp() (which is the strict \"===\") for
identifying a repeat.");
}

vecper_select(func,v) =
{
  v[1]=Vec(select(func,v[1]));
  v[2]=Vec(select(func,v[2]));
  v;
}
{
addhelp(vecper_select,
"v = vecper_select(func,v)
Return vecper v with only elements selected from it according to the given
predicate func.  func is called

    bool = func(elem)

and should return non-0 to keep the given elem.  This is simply select()
applied to the fixed and periodic parts.");
}

\\ v is a vector.
\\ If it comprises a repeating [a,b,c, a,b,c, a,b,c] then return just [a,b,c].
\\ The shortest repeating initial part is taken.
\\ If v is all 0s then return [].
vecper_INTERNAL_vector_periodic_reduce(v) =
{
  if(#v,
     fordiv(#v,len,  \\ len = prospective repeat length
           for(i=1,#v-len,
              if(v[i]!=v[i+len],
                 next(2)));  \\ not periodic of len
           return(v[1..len])));
  v;
}

vecper_normalize(v) =
{
  \\ print("vecper_normalize() ",v);
  my(pos=#(v[2]=vecper_INTERNAL_vector_periodic_reduce(v[2])));
  if(pos,
     \\ periodic part not empty
     forstep(i=#v[1],1,-1,
        \\ print("  compare i="i" pos="pos);
        if(v[1][i] != v[2][pos],
           \\ print("  different, keep 1..",i," rotate by ",#v[2]-pos);
           return([v[1][1..i],
                   vecper_INTERNAL_vector_rotate_after(v[2],pos)]));
        (pos--) || (pos=#v[2]));

     \\ print("  purely periodic, rotate after pos=",pos);
     [[], vecper_INTERNAL_vector_rotate_after(v[2],pos)];
     ,
     v);
}
{
addhelp(vecper_normalize,
"v = vecper_normalize(v)
Return vecper v in normalized form.
The period of the periodic part is reduced if it's a multiple of something
smaller.  The fixed part is reduced if some of its final terms are what the
periodic part contains anyway.  If v is purely periodic then its fixed part
is reduced to empty.

Elements are compared by \"==\" for identifying repeats.");
}


\\-----------------------------------------------------------------------------
\\ Fraction Digits

vecper_frac_digits(f,base=10,flag=0) =
{
  \\ print("vecper_frac_digits() f="f" base="base);
  base>1 || error("must have base > 1");
  (0 <= f && f <= 1) || error("must have 0 <= f <= 1");

  if(f==1, [[],[base-1]],       \\ always 9s form for f==1

    my(num,den);
    if(type(f)=="t_REAL",
       \\ binary() gives [ [bits above point], [bits below point] ]
       \\ or 0.0 is just [0,...,0]
       my(v=binary(f));
       num=if(f,subst(Pol(v[2]),'x,2)); \\ fromdigits(), but work in gp 2.7.x
       den=2^#v[2];
       ,
       \\ t_FRAC, or t_REAL 0.0
       num = numerator(f);
       den = denominator(f));

    \\ num!=0 and den==1 goes into an infinite loop building endless output
    \\ digits and num*base^k powers; so guard against that for any reason
    num==0 || den>1 || error("vecper_frac_digits() must den>1");

    my(v=vecper_make_state((num)->divrem(base*num,den),  num));
    if(v[2]==[0],
       if(flag==2,          v[2]=[],                           \\ finite form
          flag==1 && #v[1], v[1][#v[1]]--; v[2][1] = base-1)); \\ 9s form
    v);
}
{
addhelp(vecper_frac_digits,
"v = vecper_frac_digits(f,base=10,flag=0)
f is a fraction in the range 0 <= f <= 1.
Return its digits as a vecper, either decimal or the optional given \"base\".

base must be an integer >= 2.  Digits of a fraction are an initial fixed
part then periodic part so f = .ddd (eeee)repeating.

If base and denominator(f) have a common factor then normally the periodic
part is a modest length.  But if no common factor then the periodic part may
be as long as denominator(f)-1.

f can be a float (t_REAL) which is taken as a binary fraction n/2^k where n
is its binary(f) bits, and k near its precision.  If base is odd then, as
noted above, this may give a hopelessly long periodic part, up to 2^k-1.

If f is an exact f=n/base^k for integer n in the range 0 <= n < base^k, then
the periodic part is 0s, so f = .ddd (0)repeat.  Optional \"flag\" parameter
can be 1 to use trailing 9s instead f = .ddc (9)repeat.  Similarly in other
bases where \"9\" is base-1.  f=0 can only be represented by 0s and f=1 can
only be represented by 9s, so those respective forms are returned always.

\"flag\" can be 2 to represent an exact f<1 using a finite vecper, omitting
the 0s periodic part.");
}

\\ Like fromdigits(), but accepting any base.
\\ Done with a sum instead of Pol() and subst() so that the elements of v
\\ can be polynomials too, in any variable, including x.
vecper_INTERNAL_fromdigits_any(v,base) =
{
  my(ret=0);
  for(i=1,#v, ret=ret*base+v[i]);
  ret;
}

vecper_frac_fromdigits(v,base=10) =
{
  \\ print("fixed ",subst(Pol(v[1]),'x,base));
  \\ print("num ",subst(Pol(v[2]),'x,base));
  \\ print("den ",(base^length(v[2]) - 1));
  \\ print("den shift ",length(v[1]));
  \\ print("f ",subst(Pol(v[2]),'x,base) / (base^length(v[2]) - 1));

  ( vecper_INTERNAL_fromdigits_any(v[1],base)
   + if(#v[2], vecper_INTERNAL_fromdigits_any(v[2],base)/(base^#v[2] - 1)) )
  / base^#v[1];
}
{
addhelp(vecper_frac_fromdigits,
"f = vecper_frac_fromdigits(v,base=10)
v is a vecper whose elements are digits with radix point immediately above
the first vector element.  Return a fraction (t_FRAC) which is their value.

The default is decimal digits, or the optional \"base\" parameter gives
something else.  \"base\" can be any type able to be powered for

   v1/base^1 + v2/base^2 + v3/base^3 + ...

which includes complex numbers, quads, polynomials, matrices.  If base is a
polynomial 1/x then the result is like a t_RFRAC similar to vecper_polrev()
(first term coefficient of x^1).");
}

\\-----------------------------------------------------------------------------
\\  LocalWords:  vecpers func normalize normalized frac fromdigits vec len
\\  LocalWords:  vecper elem poly polrev var RFRAC bool cmp ddd eeee ddc
